import { Algebra, Monad } from './interfaces.js'
import { curry } from '../..//vendor/izuna/src/index.js'

/** @import { Applicative, ApplicativeTypeRef, Morphism } from './types.js' */


/** @import { InferredMorphism } from './types.js' */

const kleisli = curry((f, g, x) => f(x).chain(g))

/**
 * @template T
 * @type {Pure<T> | Impure<T>} Free
 */
export class Free extends Algebra(Monad) {
  /**
   * @template T
   * @param {T} value 
   */
  static of(value) {
    return new Pure(value)
  }
}

/** @template T */
export class Pure extends Free {
  #value

  /**
   * @param {T} value 
   */
  constructor(value) {
    super()
    this.#value = value
  }

  /**
   * @param {InferredMorphism<T>} f 
   */
  chain(f) {
    return f(this.#value)
  }

  /**
   * @param {InferredMorphism<T>} f 
   */
  map(f) {
    return pure(f(this.#value))
  }

  /**
   * @template U
   * @template {Applicative<U>} M
   * @template {ApplicativeTypeRef<U, M>} TypeRef
   * @param {TypeRef} _A
   * @param {*} f
   * @returns {M}
   */
  traverse(_A, f) {
    return f(this.#value).map(pure)
  }

  run() {
    return this.#value
  }
}

/** @template T */
export class Impure extends Free {
  #value
  #next

  /**
   * @param {T} value 
   * @param {(value: T) => Free<any>} f 
   */
  constructor(value, f) {
    super()
    this.#value = value
    this.#next = f
  }

  /**
   * @param {<U>(value: T) => Free<U>} f 
   */
  chain(f) {
    return impure(this.#value, kleisli(this.#next, f))
  }

  map(f) {
    return impure(
      this.#value,
      y => f(y).chain(f)
    )
  }

  /**
   * @template U
   * @template {Applicative<U>} M
   * @template {ApplicativeTypeRef<U, M>} TypeRef
   * @param {TypeRef} A
   * @param {(value: T) => M} f
   * @returns {M}
   */
  traverse(A, f) {
    const fb = f(this.#value)
    const rest = this.#next(this.#value).traverse(A, f)

    return rest.ap(
      fb.map(b => next => impure(b, () => next))
    )
  }

  run() {
    return this.#next(this.#value).run()
  }
}

/**
 * @template T
 * @param {T} x
 * @returns {Pure<T>}
 */
const pure = x => new Pure(x)

/**
 * @template T
 * @param {T} x
 * @param {(value: T) => Free<any>} f
 * @returns {Impure<T>}
 */
const impure = (x, f) => new Impure(x, f)

/**
 * @template T
 * @param {T} x
 * @returns Impure<T>
 */
export const liftF = x => impure(x, pure)
import { id, Identity } from './identity.js'
import { Option, some } from './option.js'

const a = liftF(1).chain(() => liftF(2)).traverse(Option, x => some(1))
console.log(a.chain(x => x).run())

