import { Algebra, Comonad, Monad } from './interfaces.js'

/** @import { Morphism } from './types.js' */

/** @template T */
export class Identity extends Algebra(Monad, Comonad) {
  #value

  /**
   * @param {T} value 
   */
  constructor(value) {
    super()

    this.#value = value
  }

  /**
   * @template T
   * @param {T} value 
   */
  static of(value) {
    return id(value)
  }

  /**
   * @template U
   * @param {Morphism<T, U>} f 
   */
  map(f) {
    return id(f(this.#value))
  }

  /**
   * @template U
   * @this {Identity<Morphism<U, T>>}
   * @param {Identity<U>} b 
   * @returns {Identity<T>}
   */
  ap(b) {
    return b.map(this.#value)
  }

  /**
   * @template U
   * @param {Morphism<T, Identity<U>>} f 
   */
  chain(f) {
    return f(this.#value)
  }

  /**
   * @param {(value: Identity<T>) => T} f 
   */
  extend(f) {
    return id(f(this))
  }

  extract() {
    return this.#value
  }
}

export const id = value => new Identity(value)

